#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <netdb.h>
#include <malloc.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#ifndef SUNOS
#include <sys/resource.h>
#endif /* !SUNOS */
#include <signal.h>
#include <sys/signal.h>
#include <sys/time.h>
#include <netinet/in.h>
#include "defs.h"
#include "cell.h"
#include "list.h"
#include "socket.h"
#include "etypes.h"
#include "stypes.h"
#include "mutationcenter.h"

void main(argc, argv)
    int argc;
    char **argv;
{
    int    serversocket, a_sock;
    list   *slave_list, *tmp_link, *next_link;
    slave  *curr_slave;
    int    is_first;
    arena  farena;
    char   readbuffer[SOCKET_BUFFER_SIZE];
    struct sockaddr_in addr_accept;
    int    sock_size, max_fd;
    struct timeval timeout, tv;
    struct timezone tz;
    time_t current_time;
    fd_set rd;
    struct linger  lingeropt;
    int    server_timeout = SERVER_TIMEOUT;
#ifndef SUNOS
    struct rlimit rlp;
#endif /* !SUNOS */

#ifdef SUNOS
    max_fd = getdtablesize();
#else
    getrlimit(RLIMIT_NOFILE, &rlp);
    max_fd = (int) rlp.rlim_cur;
#endif /* SUNOS */
    
    slave_list = NULL;
    sock_size=sizeof(struct sockaddr_in);
    signal(SIGPIPE, SIG_IGN);
    lingeropt.l_onoff=1;
    lingeropt.l_linger=60;
    (void)gettimeofday(&tv,&tz);
    srand((unsigned int)tv.tv_usec);
    
    serversocket = server(8007);
    listen(serversocket, 5);

    if(serversocket == -1) {
	printf("Unable to bind socket, fatal error\n");
	exit(1);
    }

    ArenaInitialize(&farena, argv, argc);
    farena.generation = 0;
    ArenaEvolve(&farena);

    while(1) {
	current_time = time((time_t *)NULL);
	FD_ZERO(&rd);
	FD_SET(serversocket, &rd);
	is_first = TRUE;
	/**
	 * go through the list, purge the timed out slaves or add
	 * the socket in the fd_set
	 */
	for(tmp_link = slave_list; tmp_link;) {
	    curr_slave = (slave *)tmp_link->cell;
	    if(difftime(current_time, curr_slave->last_time)>server_timeout) {
		printf("removing timed out client socket %d    \n",
		       curr_slave->socket);
		/**
		 * error, unregister the client and the what was processed
		 * and close the socket
		 */
		ArenaNotifyFailure(&farena, curr_slave);
		shutdown(curr_slave->socket, 2);
		close(curr_slave->socket);
		next_link = tmp_link->next;
		SwallowLink(tmp_link);
		FreeCell(CELL_SLAVE, curr_slave);
		if(is_first && !next_link)
		    slave_list = NULL;
		else
		    if(is_first)
			slave_list = next_link;
		free(tmp_link);
		tmp_link = next_link;
	    } else {
		FD_SET(curr_slave->socket, &rd);
		is_first = FALSE;
		tmp_link = tmp_link->next;
	    }
	}
	timeout.tv_sec=1;
	timeout.tv_usec=0;
	
	select(max_fd, &rd, NULL, NULL, &timeout);

	if(FD_ISSET(serversocket, &rd)) {
	    /**
	     * add the incoming connection to the list
	     */
	    a_sock = accept(serversocket, 
				  (struct sockaddr *)&addr_accept,
				  &sock_size);
	    if(a_sock != -1)
		AddCellFirst(&slave_list, CELL_SLAVE, CreateSlave(a_sock)); 
	    printf("adding client socket %d       \n", a_sock);
	}
	
	is_first = TRUE;
	for(tmp_link = slave_list; tmp_link;) {
	    /**
	     * i should check if tmp_link->type == CELL_SLAVE
	     * but it can't happen ;)
	     */
	    curr_slave = (slave *)tmp_link->cell;
	    if(FD_ISSET(curr_slave->socket, &rd)) {
		curr_slave->last_time = current_time;
		curr_slave->status = read_socket(curr_slave->socket, 
						 curr_slave->buffer,
						 &(curr_slave->curr),
						 readbuffer);
		if(curr_slave->status == 1) {
		    ArenaParse(&farena, curr_slave, readbuffer);
		    is_first = FALSE;
		    tmp_link = tmp_link->next;
		} else {
		    /**
		     * socket reading error, close everything and update
		     */
		    printf("removing client socket %d      \n",
			   curr_slave->socket);
		    ArenaNotifyFailure(&farena, curr_slave);
		    shutdown(curr_slave->socket, 2);
		    close(curr_slave->socket);
		    next_link = tmp_link->next;
		    SwallowLink(tmp_link);
		    FreeCell(CELL_SLAVE, curr_slave);
		    if(is_first && !next_link)
			slave_list = NULL;
		    else
			if(is_first)
			    slave_list = next_link;
		    free(tmp_link);
		    tmp_link = next_link;
		} 
	    } else {
		is_first = FALSE;
		tmp_link = tmp_link->next;
	    }
	}
	ArenaCheckPending(&farena, &slave_list);
    }
}
