#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include "types.h"
#include "tools.h"
#include "strategie.h"
#include "brain.h"
#include "mutatedefs.h"
#include "mutate.h"

void MutateBrainCellFromSeed(seed, brain)
    param_cell seed, *brain;
{
    brain->my_score_rate    = seed.my_score_rate    + (rand()%128)/MSR - MSRD;
    brain->opp_score_rate   = seed.opp_score_rate   + (rand()%128)/OSR - OSRD;
    brain->protect_pos_rate = seed.protect_pos_rate + (rand()%128)/PPR - PPRD;
    brain->attack_pos_rate  = seed.attack_pos_rate  + (rand()%128)/APR - APRD;
    brain->board_score_rate = seed.board_score_rate + (rand()%128)/BSR - BSRD;
    brain->play_opp_flag    = seed.play_opp_flag    + (rand()%128)/POF - POFD;
    brain->defensive_rate   = seed.defensive_rate   + (rand()%128)/DR  - DRD;
    brain->offensive_rate   = seed.offensive_rate   + (rand()%128)/OR  - ORD;
    brain->position_rate    = seed.position_rate    + (rand()%128)/PR  - PRD;
}

void MutateBrainFromSeed(seed, brain)
    param seed, brain;
{
    int i;
    
    for(i=0;i<42;i++)
	MutateBrainCellFromSeed(seed[i], &brain[i]);
}

void MutateAddBrainCell(seed, brain, scale)
    param_cell seed, *brain;
    PARAM_TYPE scale;
{
    brain->my_score_rate    += seed.my_score_rate    / scale;
    brain->opp_score_rate   += seed.opp_score_rate   / scale;
    brain->protect_pos_rate += seed.protect_pos_rate / scale;
    brain->attack_pos_rate  += seed.attack_pos_rate  / scale;
    brain->board_score_rate += seed.board_score_rate / scale;
    brain->play_opp_flag    += seed.play_opp_flag    / scale;
    brain->defensive_rate   += seed.defensive_rate   / scale;
    brain->offensive_rate   += seed.offensive_rate   / scale;
    brain->position_rate    += seed.position_rate    / scale;
}  

void MutateAddBrain(seed, brain, scale)
    param seed, brain;
    PARAM_TYPE scale;
{
    int i;
    
    for(i=0;i<42;i++)
	MutateAddBrainCell(seed[i], &brain[i], scale);
}

void SplitBrainCell(seed1, seed2, brain, num)
    param_cell seed1, seed2, *brain;
    int num;
{
    brain->my_score_rate    = (seed1.my_score_rate    * num + seed2.my_score_rate    * (41-num)) / 41;
    brain->opp_score_rate   = (seed1.opp_score_rate   * num + seed2.opp_score_rate   * (41-num)) / 41;
    brain->protect_pos_rate = (seed1.protect_pos_rate * num + seed2.protect_pos_rate * (41-num)) / 41;
    brain->attack_pos_rate  = (seed1.attack_pos_rate  * num + seed2.attack_pos_rate  * (41-num)) / 41;
    brain->board_score_rate = (seed1.board_score_rate * num + seed2.board_score_rate * (41-num)) / 41;
    brain->play_opp_flag    = (seed1.play_opp_flag    * num + seed2.play_opp_flag    * (41-num)) / 41;
    brain->defensive_rate   = (seed1.defensive_rate   * num + seed2.defensive_rate   * (41-num)) / 41;
    brain->offensive_rate   = (seed1.offensive_rate   * num + seed2.offensive_rate   * (41-num)) / 41;
    brain->position_rate    = (seed1.position_rate    * num + seed2.position_rate    * (41-num)) / 41;
}

void SplitBrain(seed1, seed2, brain)
    param seed1, seed2, brain;
{
    int i;
    
    for(i=0;i<42;i++)
	SplitBrainCell(seed1[i], seed2[i], &brain[i], i);
}

void MixBrain(seed1, seed2, brain)
    param seed1, seed2, brain;
{
    int i;
    
    for(i=0;i<42;i++)
	SplitBrainCell(seed1[i], seed2[i], &brain[i], 21);
}
    
void MutateAddGeneBrainCell(seed, brain, gene)
    param_cell seed, *brain;
    int gene;
{
    switch(gene) {
    case 1:
	brain->my_score_rate    = seed.my_score_rate    + (rand()%128)/MSR - MSRD;
	break;
    case 2:
	brain->opp_score_rate   = seed.opp_score_rate   + (rand()%128)/OSR - OSRD;
	break;
    case 3:
	brain->protect_pos_rate = seed.protect_pos_rate + (rand()%128)/PPR - PPRD;
	break;
    case 4:
	brain->attack_pos_rate  = seed.attack_pos_rate  + (rand()%128)/APR - APRD;
	break;
    case 5:
	brain->board_score_rate = seed.board_score_rate + (rand()%128)/BSR - BSRD;
	break;
    case 6:
	brain->play_opp_flag    = seed.play_opp_flag    + (rand()%128)/POF - POFD;
	break;
    case 7:
	brain->defensive_rate   = seed.defensive_rate   + (rand()%128)/DR  - DRD;
	break;
    case 8:
	brain->offensive_rate   = seed.offensive_rate   + (rand()%128)/OR  - ORD;
	break;
    default:
	brain->position_rate    = seed.position_rate    + (rand()%128)/PR  - PRD;
	break;
    }
}

void MutateAddGeneBrain(seed, brain, gene)
    param seed, brain;
    int gene;
{
    int i;
    
    for(i=0;i<42;i++)
	MutateAddGeneBrainCell(seed[i], &brain[i], gene);
}
 
void MutateMulGeneBrainCell(seed, brain, gene, factor)
    param_cell seed, *brain;
    int gene;
    float factor;
{
    switch(gene) {
    case 1:
	brain->my_score_rate    = seed.my_score_rate    * factor;
	break;
    case 2:
	brain->opp_score_rate   = seed.opp_score_rate   * factor;
	break;
    case 3:
	brain->protect_pos_rate = seed.protect_pos_rate * factor;
	break;
    case 4:
	brain->attack_pos_rate  = seed.attack_pos_rate  * factor;
	break;
    case 5:
	brain->board_score_rate = seed.board_score_rate * factor;
	break;
    case 6:
	brain->play_opp_flag    = seed.play_opp_flag    * factor;
	break;
    case 7:
	brain->defensive_rate   = seed.defensive_rate   * factor;
	break;
    case 8:
	brain->offensive_rate   = seed.offensive_rate   * factor;
	break;
    default:
	brain->position_rate    = seed.position_rate    * factor;
	break;
    }
}

void MutateMulGeneBrain(seed, brain, gene, factor)
    param seed, brain;
    int gene;
    float factor;
{
    int i;
    
    for(i=0;i<42;i++)
	MutateMulGeneBrainCell(seed[i], &brain[i], gene, factor);
}
 
void MutateAggressivenessBrain(seed, brain, factor)
    param seed, brain;
    float factor;
{
    MutateMulGeneBrain(seed, brain, 4, factor);
    MutateMulGeneBrain(seed, brain, 3, 1/factor);
    MutateMulGeneBrain(seed, brain, 8, factor);
    MutateMulGeneBrain(seed, brain, 7, 1/factor);
}

void MutateExpansivenessBrain(seed, brain, factor)
    param seed, brain;
    float factor;
{
    MutateMulGeneBrain(seed, brain, 1, factor);
    MutateMulGeneBrain(seed, brain, 2, 1/factor);
    MutateMulGeneBrain(seed, brain, 5, factor);
    MutateMulGeneBrain(seed, brain, 9, factor);
}
