#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <sys/socket.h>
#include "defs.h"
#include "types.h"
#include "tools.h"
#include "strategie.h"
#include "brain.h"
#include "mutate.h"
#include "list.h"
#include "cell.h"
#include "scell.h"
#include "etypes.h"
#include "stypes.h"
#include "mutationcenter.h"
#include "socket.h"
#ifndef WIN_SCORE
#   define WIN_SCORE 3
#endif
#ifndef NULL_SCORE
#   define NULL_SCORE 0
#endif
#ifndef LOSS_SCORE
#   define LOSS_SCORE -1
#endif

#ifdef LOG
	FILE *lf;
#endif

void BubbleBrain(btab, stab)
    param btab[MAX_SIZE];
    int stab[MAX_SIZE][MAX_SIZE];
{
    int sres[MAX_SIZE];
    int i,j, idump,ok;
    param bdump;

    for(i=0; i<MAX_SIZE;i++) {
	sres[i] = 0;
	for(j=0;j<MAX_SIZE;j++)
	    if(i!=j) {
#ifdef WIN
		sres[i] = sres[i] + stab[i][j];
		sres[i] += (stab[j][i] > 0) ? LOSS_SCORE :
			((stab[j][i])? WIN_SCORE : NULL_SCORE);
#else /* use score only */
		sres[i] = sres[i] + stab[i][j] - stab[j][i];
#endif /* WIN */
	    }
#ifdef MDEBUG
	printf("[%d]", sres[i]);
#endif /* MDEBUG */
    }
#ifdef MDEBUG
	printf("\n");
#endif /* MDEBUG */   

    ok=0;
    while(!ok) {
	ok=1;
	for(i=8;i<127;i++) {
	    if(sres[i]<sres[i+1]) {
		idump = sres[i];
		sres[i] = sres[i+1];
		sres[i+1]=idump;
		CopyBrain(btab[i], bdump);
		CopyBrain(btab[i+1], btab[i]);
		CopyBrain(bdump, btab[i+1]); 
		ok = 0;
	    }
	}
    }
#ifdef MDEBUG
    for(i=0; i<MAX_SIZE;i++) 
	printf("[%d]", sres[i]);
    printf("\n");
#endif /* MDEBUG */
}

void EvolvePool(btab, stab)
    param btab[MAX_SIZE];
    int stab[MAX_SIZE][MAX_SIZE];
{
    int i;

    BubbleBrain(btab, stab);

    for(i=0; i<8;i++) {
	MutateAggressivenessBrain(btab[i+8], btab[i+16], 0.77);
	MutateExpansivenessBrain(btab[i+8], btab[i+24], 1.3);
	MutateExpansivenessBrain(btab[i+8], btab[i+32], 0.77);
	SplitBrain(btab[rand()%8+8], btab[rand()%8+8], btab[i+40]);
	MixBrain(btab[rand()%8+8], btab[rand()%8+8], btab[i+48]);
	/* breed with some braindeads and original seeds */
	MixBrain(btab[rand()%8], btab[127-rand()%8], btab[i+56]);
	MutateBrainFromSeed(btab[i+8], btab[i+64]);
	MutateAddGeneBrain(btab[i+8], btab[i+72], rand()%9);
	MutateAddGeneBrain(btab[i+8], btab[i+80], rand()%9);
	MutateMulGeneBrain(btab[i+8], btab[i+88], 1.3, rand()%9);
	MutateMulGeneBrain(btab[i+8], btab[i+96], 0.77, rand()%9);
	MixBrain(btab[i+8*(rand()%13)], btab[i+8*(rand()%13)], btab[i+104]);
	SplitBrain(btab[i+8*(rand()%13)], btab[i+8*(rand()%13)], 
		   btab[i+104]);
	MutateAggressivenessBrain(btab[i+8], btab[i+112], 1.3);
    }
}
 
void ArenaInit (ai)
    arena *ai;
{
    int i,j;

    ai->nb_played = 0;
    for(i=0; i<MAX_SIZE; i++)
	for(j=0; j<MAX_SIZE; j++) {
	    ai->arena_win[i][j]     = 0;
	    ai->arena_resultx[i][j] = 0;
	    ai->arena_resulty[i][j] = 0;
	}
}

void ArenaCheckPending(ai, sl)
    arena *ai;
    list  **sl;
{
    int        is_first;
    list       *tmp_link, *next_link, *tmp_ai;
    slave      *curr_slave;
    arena_cell *curr_arena_cell;
    
    /* are we finished with the current generation? */
    
    if(!ai->pending && !ai->waiting) {
	ArenaEvolve(ai);
    }
    
    /* put some load to the slaves, if things are to be done */
    if(ai->pending) {
	is_first = TRUE;
	for(tmp_link = *sl; tmp_link && ai->pending;) {
	    curr_arena_cell = (arena_cell *)ai->pending->cell;
	    curr_slave = (slave *)tmp_link->cell;
	    if(!curr_slave->used) {
		if(SlaveSendRequest(ai, curr_slave, curr_arena_cell->x,
				    curr_arena_cell->y)) {
		    curr_slave->used = 1;
		    curr_slave->opp_x = curr_arena_cell->x;
		    curr_slave->opp_y = curr_arena_cell->y;
		    AddCellLast(&ai->waiting, CELL_ARENA, CreateArenaCell(
			curr_arena_cell->x, curr_arena_cell->y));
		    tmp_ai = ai->pending;
		    ai->pending = ai->pending->next;
		    if(ai->pending)
			ai->pending->prev = NULL;
		    free(tmp_ai);
		    FreeCell(CELL_ARENA, curr_arena_cell);
		    tmp_link = tmp_link->next;
		    is_first = FALSE;
		} else {
		    printf("closing client socket %d          \n",
			   curr_slave->socket);
		    shutdown(curr_slave->socket, 2);
		    close(curr_slave->socket);
		    next_link = tmp_link->next;
		    SwallowLink(tmp_link);
		    FreeCell(CELL_SLAVE, curr_slave);
		    if(is_first && !next_link)
			*sl = NULL;
		    else
			if(is_first)
			    *sl = next_link;
		    free(tmp_link);
		    tmp_link = next_link;
		}
	    } else {
		tmp_link = tmp_link->next;
		is_first = FALSE;
	    }
	}
    }
}

int SlaveSendRequest(ai, sl, x, y)
    arena *ai;
    slave *sl;
    int x,y;
{
    char dump[512];
    int i, ok;
    param *br1, *br2;

    br1 = &ai->brains[x];
    br2 = &ai->brains[y];

#ifdef DEBUG
    printf("Sending (%d/%d) to socket %d\n", x, y, sl->socket);
#endif /* DEBUG */
    for(i=0, ok=1;ok && i<42; i++) {
	sprintf(dump, "%f %f %f %f %f %f %f %f %f\n", 
		(*br1)[i].my_score_rate,
		(*br1)[i].opp_score_rate,
		(*br1)[i].protect_pos_rate,
		(*br1)[i].attack_pos_rate,
		(*br1)[i].board_score_rate,
		(*br1)[i].play_opp_flag,
		(*br1)[i].defensive_rate,
		(*br1)[i].offensive_rate,
		(*br1)[i].position_rate);
	ok = send_socket(sl->socket, dump);
    }
    if(ok) {
	for(i=0; ok && i<42; i++) {
	    sprintf(dump, "%f %f %f %f %f %f %f %f %f\n", 
		    (*br2)[i].my_score_rate,
		    (*br2)[i].opp_score_rate,
		    (*br2)[i].protect_pos_rate,
		    (*br2)[i].attack_pos_rate,
		    (*br2)[i].board_score_rate,
		    (*br2)[i].play_opp_flag,
		    (*br2)[i].defensive_rate,
		    (*br2)[i].offensive_rate,
		    (*br2)[i].position_rate);
	    ok = send_socket(sl->socket, dump);
	}
    }
    if(ok) {
	sl->used = 1;
	sl->opp_x = x;
	sl->opp_y = y;
    }
    return ok;
}

void ArenaEvolve(ai)
    arena *ai;
{
    int i,j;
#ifdef LOG
    char fname[64];
#endif /* LOG */
    EvolvePool(ai->brains, ai->arena_win);
    ArenaInit(ai);
    SaveBrain("center0.brn", ai->brains[8]);
    SaveBrain("center1.brn", ai->brains[9]);
    SaveBrain("center2.brn", ai->brains[10]);
    SaveBrain("center3.brn", ai->brains[11]);
    SaveBrain("center4.brn", ai->brains[12]);
    SaveBrain("center5.brn", ai->brains[13]);
    SaveBrain("center6.brn", ai->brains[14]);
    SaveBrain("center7.brn", ai->brains[15]);
    for(i=0; i<MAX_SIZE; i++)
	for(j=0; j<MAX_SIZE; j++) 
	    AddCellFirst(&ai->pending, CELL_ARENA, CreateArenaCell(i, j));
    ai->generation++;
#ifdef LOG
    sprintf(fname, "center-playlog-generation-%d.log", ai->generation);
    fclose(lf);
    lf = fopen(fname, "a");
#endif /* LOG */
    printf("Building generation %d                   \n", ai->generation);
}

void ArenaParse(ai, sl, tobeparsed)
    arena *ai;
    slave *sl;
    char *tobeparsed;
{
    int  x_sc, y_sc;
    char dump[1024]= "";
    char log[1024]="";

    if(sl->status) {
	sscanf(tobeparsed, "%s %d %d %s", log, &x_sc, &y_sc, dump);
	if(!strcmp(dump, "END")) {
	    ai->arena_resultx[sl->opp_x][sl->opp_y] = x_sc;
	    ai->arena_resulty[sl->opp_x][sl->opp_y] = y_sc;
	    /* remove from the waiting queue */
	    PurgeCell(&ai->waiting, sl->opp_x, sl->opp_y);
#ifdef WIN
	    if(x_sc > y_sc)
		ai->arena_win[sl->opp_x][sl->opp_y] = WIN_SCORE;
	    else
		if(x_sc < y_sc)
		    ai->arena_win[sl->opp_x][sl->opp_y] = LOSS_SCORE;
		else
		    ai->arena_win[sl->opp_x][sl->opp_y] = NULL_SCORE;
#else /*  use score instead */
	   ai->arena_win[sl->opp_x][sl->opp_y] = x_sc - y_sc;
#endif /* WIN */
	    sl->used = 0;
#ifdef DEBUG
	    printf("notify end [%d] %s (%d vs %d) %d/%d\n", sl->socket,
		   log, sl->opp_x, sl->opp_y, x_sc, y_sc);
#endif /* DEBUG */
#ifdef LOG
	    fprintf(lf, "%s %d %d\n", log, x_sc, y_sc);
#endif /* LOG */
	    ++ai->nb_played;
#ifdef VERBOSE
	    printf("-> played so far: %f%c\n\E[A", 
		   ((float)ai->nb_played)/163.84, 37);
#endif /* VERBOSE */
	}
    }
}

void ArenaNotifyFailure(ai, sl)
    arena *ai;
    slave *sl;
{
    /**
     * failure, revert the cell from ai->waiting to ai->pending
     */
    PurgeCell(&ai->waiting, sl->opp_x, sl->opp_y);
    AddCellFirst(&ai->pending, CELL_ARENA, 
		 CreateArenaCell(sl->opp_x, sl->opp_y));
}

void ArenaInitialize(ai, files, num)
    arena *ai;
    char  **files;
    int num;
{
    int i, j, k, l;
    ai->waiting = NULL;
    ai->pending = NULL;
    ArenaInit(ai);
#ifdef LOG
    lf = fopen("center.started","w");
#endif /* LOG */	 
    for(i=1; i<num; i++) {
	LoadBrain(files[i], ai->brains[i-1]);
	printf("Loading brain %s\n", files[i]);
    }
    if(!(num>1))
	LoadBrain("seed1.brn", ai->brains[0]);
    
    if(!(num > 2))
	LoadBrain("seed2.brn", ai->brains[1]);
    ai->generation = 0;
    if(!(num > 2))
	k=2;
    else
	k=num-1;
    i=0;
    l=1;
    for(j=k;j<128;j++) {
	if(j%2) {
	    SplitBrain(ai->brains[l], ai->brains[i], ai->brains[j]);
	    i++;
	    if(i>k) {
		i=0;
		l++;
	    }
	}
	else
	    MutateBrainFromSeed(ai->brains[j%k], ai->brains[j]);
    }
}
