#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include "types.h"
#include "tools.h"
#include "strategie.h"
#include "brain.h"
#include "mutate.h"

#define PMIN(a,b) ((a<b)?a:b)
#define MAX_TURN 19

void BubbleBrain(param[32], int[32][32]);
void EvolvePool(param[32], int[32][32]);

void BubbleBrain(btab, stab)
    param btab[32];
    int stab[32][32];
{
    int sres[32];
    int i,j, idump,ok;
    param bdump;

    for(i=1; i<31;i++)
    {
	sres[i] = 0;
	for(j=0;j<32;j++)
	    if(i!=j)
		sres[i] = sres[i] + stab[i][j] - stab[j][i];
    }
    
    ok=0;
    while(!ok)
    {
	ok=1;
	for(i=1;i<30;i++)
	{
	    if(sres[i]<sres[i+1])
	    {
		idump = sres[i];
		sres[i] = sres[i+1];
		sres[i+1]=idump;
		CopyBrain(btab[i], bdump);
		CopyBrain(btab[i+1], btab[i]);
		CopyBrain(bdump, btab[i+1]); 
		ok = 0;
	    }
	}
    }
}

void EvolvePool(btab, stab)
    param btab[32];
    int stab[32][32];
{
    int i;

    BubbleBrain(btab, stab);
 
    /*  
     *  We keep the seeds (0 and 31)
     *  We keep the 7 best (1,2,3,4,5,6,7)
     *  We mutate (8,9,10,11,12,13,14) from seed 0
     *  We mutate 15 as a mean brain between the best (1..7) ones.
     *  We mutate (16,17,18,19,20,21,22) from seed 'best one' aka brain 1
     *  We mutate 23 as a mean brain between the best (1..7) and the two seeds (0, 31)
     *  We mutate (24,25,26,27,28,29,30) from seed 31
     */
 
    memset(&btab[15], 0, sizeof(param));
    memset(&btab[23], 0, sizeof(param));
    for(i=1;i<8;i++)
    {
	MutateAddBrain(btab[i], btab[15], 7);
	MutateAddBrain(btab[i], btab[23], 9);
    }
    MutateAddBrain(btab[0], btab[23], 9);
    MutateAddBrain(btab[31], btab[23], 9);

    for(i=8;i<15;i++)
	MutateBrainFromSeed(btab[0], btab[i]);

    for(i=16;i<23;i++)
	MutateBrainFromSeed(btab[1], btab[i]);

    for(i=24;i<31;i++)
	MutateBrainFromSeed(btab[31], btab[i]);
}
    
void main(argc, argv)
    int argc;
    char **argv;
{
    char *dumpname;
    board pb, cb;
    int i, nb, c, pl, result, j, k;
    pred pwhere;
    param brain[32];
    int results[32][32];
    int generation;
    struct timeval tv;
    struct timezone tz;

    (void)gettimeofday(&tv,&tz);
    srand((unsigned int)tv.tv_usec);
   
    if(argc>1)
	LoadBrain(argv[1], brain[0]);
    else
	LoadBrain("seed1.brn", brain[0]);
   
    if(argc > 2)
	LoadBrain(argv[2], brain[31]);
    else
	LoadBrain("seed2.brn", brain[31]);

    dumpname = (argc == 4) ? argv[3] : NULL;
    nb = 2;

    generation = 0;

    for(i=1;i<15;i++)
	MutateBrainFromSeed(brain[0], brain[i]);
    
    SplitBrain(brain[0], brain[31], brain[15]);
    SplitBrain(brain[31], brain[0], brain[16]);
    
    for(i=17;i<31;i++)
	MutateBrainFromSeed(brain[31], brain[i]);

    while(1)
    {
	for(j=0;j<32;j++)
	{
	    for(k=0; k<32; k++)
	    {
		if(j!=k)
		{
		    for(i=0;i<36;i++)
		    {
			pb[i] = cb[i] = 0;
		    }
		    pl = 1;
		    for(i=0;i<MAX_TURN;i++)
		    {
			c = AlphaBeta(pb, cb, pl, nb, i*2, pwhere, brain[j]);
			if(pl>0)
			    PlayWhite(pb, cb, pwhere[nb]);
			else
			    PlayBlack(pb, cb, pwhere[nb]);
			pl = -pl;
			c = AlphaBeta(pb, cb, pl, nb, i*2+1, pwhere, brain[k]);
			if(pl>0)
			    PlayWhite(pb, cb, pwhere[nb]);
			else
			    PlayBlack(pb, cb, pwhere[nb]);
			pl = -pl;
		    }
		    for(i=2*MAX_TURN;i<42;i++)
		    {
			c = EndAlphaBeta(pb, cb, pl, PMIN(4, 42-i), i, pwhere);
			if(pl>0)
			    PlayWhite(pb, cb, pwhere[PMIN(4, 42-i)]);
			else
			    PlayBlack(pb, cb, pwhere[PMIN(4, 42-i)]);
			pl = -pl;
		    }
		    result = EndScore(pb, cb, 1);
		    if(result)
			if(result>0)
			    results[j][k] = 1;
			else
			    results[j][k] = -1;
		    else
			results[j][k] = 0;
		}
	    }
	}
	printf("\nGENERATION %d\n", generation);
	for(j=0;j<32;j++)
	{
	    for(k=0;k<32;k++)
	    {
		if(j == k)
		    printf(" X ");
		else
		    if(results[j][k]<0)
			printf(" - ");
		    else
			if(results[j][k]>0)
			    printf(" 1 ");
			else
			    printf(" = ");
	    }
	    printf("\n");   
	}
	EvolvePool(brain, results);
	SaveBrain((dumpname) ? dumpname : "firstbrain.brn", brain[1]);
	generation++;
    }
}
