#ifdef VERBOSE_ROUTE
#include <stdio.h>
#endif /* VERBOSE_ROUTE */
#ifdef ADD_RANDOM
#include <stdlib.h>
#endif /* ADD_RANDOM */
#include "types.h"
#include "tools.h"
#include "strategie.h"
#ifndef HARDCUT
#define HARDCUT 8
#define IHARDCUT 36
#endif
#ifndef STEP
#define STEP 2
#endif

     /*
      * some static proto
      */

static PARAM_TYPE BoardScore(board, board, int, param_cell);
static PARAM_TYPE StrategicScore(board, board, int, int, param_cell);
static PARAM_TYPE SearchScore(board, board, PARAM_TYPE, PARAM_TYPE, int, param_cell);
static void Shuffler     (board , board , int  , int,   int *, param);  
static void EndShuffler  (board , board , int  , int *);
static void FastEndShuffler  (board , board , int  , int *);
static PARAM_TYPE PAlphaBetaW (PARAM_TYPE, PARAM_TYPE, board, board, int ,int , pred, param, PARAM_TYPE, PARAM_TYPE);
static PARAM_TYPE PAlphaBetaB (PARAM_TYPE, PARAM_TYPE, board, board, int ,int , pred, param, PARAM_TYPE, PARAM_TYPE);
static int  PEndAlphaBetaB (board , board , int  , int  , pred, pred, int, int);
static int  PEndAlphaBetaW (board , board , int  , int  , pred, pred, int, int);
static int  PFastEndAlphaBetaB (board , board , int  , int  , pred, pred, int, int);
static int  PFastEndAlphaBetaW (board , board , int  , int  , pred, pred, int, int);

    /*
     * Explanations about AB schemes used, evalutation...
     */

static PARAM_TYPE BoardScore(play_b, calc_b, player, pcell)
    board play_b, calc_b;
    int player;
    param_cell pcell;
{
    register int dump;
    int i;
    PARAM_TYPE score;
    
    score = 0;
    if(player>0) /* White player */
    {
	for(i=0;i<36;i++)
	{
	    dump = play_b[i]+calc_b[i];
	    if(dump > 0)
		score += pcell.my_score_rate;
	    else
		if(dump < 0)
		    score -= pcell.opp_score_rate;
	    
	    if(calc_b[i] == 0)
		if(play_b[i]>0)
		    score -= pcell.protect_pos_rate;
		else
		    if(play_b[i]<0)
			score += pcell.attack_pos_rate;
	} 
    }
    else
    {                    /* Black Player */
	for(i=0;i<36;i++)
	{
	    dump = play_b[i]+calc_b[i];
	    if(dump < 0)
		score += pcell.my_score_rate;
	    else
		if(dump > 0)
		    score -= pcell.opp_score_rate;
	    
	    if(calc_b[i] == 0)
		if(play_b[i]<0)
		    score -= pcell.protect_pos_rate;
		else
		    if(play_b[i]>0)
			score += pcell.attack_pos_rate;
	}
    }
    return score;
}

static PARAM_TYPE StrategicScore(oplay_b, ocalc_b, player, where, pcell)
    board oplay_b, ocalc_b;
    int player, where;
    param_cell pcell;
{
#define WTestDefensive(x) if(oplay_b[x]>0 && ocalc_b[x] == 0) weights += pcell.defensive_rate
#define WTestOffensive(x) if(oplay_b[x]<0 && ocalc_b[x] == 0) weights += pcell.offensive_rate
#define BTestDefensive(x) if(oplay_b[x]>0 && ocalc_b[x] == 0) weights += pcell.offensive_rate
#define BTestOffensive(x) if(oplay_b[x]<0 && ocalc_b[x] == 0) weights += pcell.defensive_rate
    
    PARAM_TYPE weights;

    weights = 0;
    if(player>0) /* White player */
    {
		if(ocalc_b[where]<0)
	    weights -= pcell.play_opp_flag;

        WTestDefensive(where);
	if(where>5)
	{
	    WTestDefensive(where-6);
	    WTestOffensive(where-6);
	}
	else
	    weights -= pcell.position_rate;
	if(where<30)
	{
	    WTestDefensive(where+6);
	    WTestOffensive(where+6);
	}
	else
	    weights -= pcell.position_rate;
	if(where != 0 && where != 6 && where != 12 && where != 18 && where != 24 && where !=30)
	{
	    WTestDefensive(where-1);
	    WTestOffensive(where-1);
	}
	else
	    weights -= pcell.position_rate;
	if(where != 5 && where != 11 && where != 17 && where != 23 && where != 29 && where != 35)
	{
	    WTestDefensive(where+1);
	    WTestOffensive(where+1);
	}
	else
	    weights -= pcell.position_rate;
    }
    else
    {                    /* Black Player */
	if(ocalc_b[where]>0)
	    weights -= pcell.play_opp_flag;
	
        BTestDefensive(where);
	if(where>5)
	{
	    BTestDefensive(where-6);
	    BTestOffensive(where-6);
	}
	else
	    weights -= pcell.position_rate;
	if(where<30)
	{
	    BTestDefensive(where+6);
	    BTestOffensive(where+6);
	}
	else
	    weights -= pcell.position_rate;
	if(where != 0 && where != 6 && where != 12 && where != 18 && where != 24 && where !=30)
	{
	    BTestDefensive(where-1);
	    BTestOffensive(where-1);
	}
	else
	    weights -= pcell.position_rate;
	if(where != 5 && where != 11 && where != 17 && where != 23 && where != 29 && where != 35)
	{
	    BTestDefensive(where+1);
	    BTestOffensive(where+1);
	}
	else
	    weights -= pcell.position_rate;
    }	
    return weights;
}

static PARAM_TYPE SearchScore(play_b, calc_b, sc_weight, st_weight, player, pcell)
    board play_b, calc_b;
    PARAM_TYPE sc_weight, st_weight;
    int player;
    param_cell pcell;
{
    PARAM_TYPE score;
#ifdef NO_ADD_SCORE
    PARAM_TYPE b_score;
#endif /* NO_ADD_SCORE */

    score = BoardScore(play_b, calc_b, player, pcell) * sc_weight;

#ifndef NO_ADD_SCORE
    score += st_weight;
#else
    if(score<0)
	b_score = 1 - score;
    else
	b_score = score + 1;

    score += b_score * st_weight;
#endif /* NO_ADD_SCORE */

#ifdef ADD_RANDOM
    score = score*8 + (rand()&7);
#endif /* ADD_RANDOM */
    return score;
}

static void Shuffler(play_b, calc_b, player, num_coup, where_list, brain)
    board play_b, calc_b;
    int player, num_coup, where_list[36];
    param brain;
{
    board d2_play_b, d1_play_b;
    board d2_calc_b, d1_calc_b;
    PARAM_TYPE alpha, beta, dump, st_weight, sc_weight;
    int pris_d1, pris_d2, ok;
    int where_d1, where_d2, where_b;
    int where_result[36];
    
    alpha = -1000000;
    where_b = -1;
    sc_weight = brain[num_coup].board_score_rate;

    Copy(play_b, d1_play_b);
    Copy(calc_b, d1_calc_b);

    for(where_d1=0; where_d1<36; where_d1++)
    {
	where_list[where_d1] = where_d1;
	where_result[where_d1] = -1000000;
    }

    if(player>0)	/* White */
    {
	for(where_d1=0; where_d1<36; where_d1++)
	{
	    ok=1;
	    beta = 1000000;
	    if(IsValidMove(d1_play_b, player, where_d1))
	    {
		st_weight = StrategicScore(play_b, calc_b, player, where_d1, brain[num_coup]);		
		pris_d1 = PlayWhite(d1_play_b, d1_calc_b, where_d1);
		Copy(d1_play_b, d2_play_b);
		Copy(d1_calc_b, d2_calc_b);
		for(where_d2=0; ok && where_d2<36; where_d2++)
		{
		    if(IsValidMove(d2_play_b, -1, where_d2))
		    {
			pris_d2 = PlayBlack(d2_play_b, d2_calc_b, where_d2);
			dump = SearchScore(d2_play_b, d2_calc_b, sc_weight, st_weight, 1, brain[num_coup+1]);
			if(dump<beta)
			{
			    beta = dump;
			    if(dump <= alpha)
				ok = 0;
			}
			if(pris_d2)
			{
			    Copy(d1_play_b, d2_play_b);
			    Copy(d1_calc_b, d2_calc_b);
			}
			else
			    UnPlayBlack(d2_play_b, d2_calc_b, where_d2);
		    }
		}
		if (beta>alpha)
		{
		    alpha = beta;
		    where_b = where_d1;
		}
		if(pris_d1)
		{
		    Copy(play_b, d1_play_b);
		    Copy(calc_b, d1_calc_b);
		}
		else
		    UnPlayWhite(d1_play_b, d1_calc_b, where_d1);
		where_result[where_d1] = beta;
	    }
	    else
		where_result[where_d1] = -100000;
	}
    }
    else		/* Black */
    {
	for(where_d1=0; where_d1<36; where_d1++)
	{
	    ok=1;
	    beta = 1000000;
	    if(IsValidMove(d1_play_b, player, where_d1))
	    {
		st_weight = StrategicScore(play_b, calc_b, player, where_d1, brain[num_coup]);
		pris_d1 = PlayBlack(d1_play_b, d1_calc_b, where_d1);
		Copy(d1_play_b, d2_play_b);
		Copy(d1_calc_b, d2_calc_b);
		for(where_d2=0; ok && where_d2<36; where_d2++)
		{
		    if(IsValidMove(d2_play_b, 1, where_d2))
		    {
			pris_d2 = PlayWhite(d2_play_b, d2_calc_b, where_d2);
			dump = SearchScore(d2_play_b, d2_calc_b, sc_weight, st_weight, -1, brain[num_coup+1]);
			if(dump<beta)
			{
			    beta = dump;
			    if(dump <= alpha)
				ok = 0;
			}
			if(pris_d2)
			{
			    Copy(d1_play_b, d2_play_b);
			    Copy(d1_calc_b, d2_calc_b);
			}
			else
			    UnPlayWhite(d2_play_b, d2_calc_b, where_d2);
		    }
		}
		if (beta>alpha)
		{
		    alpha = beta;
		    where_b = where_d1;
		}
		if(pris_d1)
		{
		    Copy(play_b, d1_play_b);
		    Copy(calc_b, d1_calc_b);
		}
		else
		    UnPlayBlack(d1_play_b, d1_calc_b, where_d1);  
		where_result[where_d1] = beta;
	    }
	    else
		where_result[where_d1] = -100000;
	}
    }
    BubbleSort(where_list, where_result);
}

static void EndShuffler(play_b, calc_b, player, where_list)
    board play_b, calc_b;
    int player, where_list[36];
{
    board d2_play_b, d1_play_b;
    board d2_calc_b, d1_calc_b;
    int alpha, beta;
    int pris_d1, pris_d2, ok, dump;
    int where_d1, where_d2, where_b;
    int where_result[36];

    alpha = -1000000;
    where_b = -1;

    Copy(play_b, d1_play_b);
    Copy(calc_b, d1_calc_b);

    for(where_d1=0; where_d1<36; where_d1++)
    {
	where_list[where_d1] = where_d1;
	where_result[where_d1] = -1000000;
    }

    if(player>0)	/* White */
    {
	for(where_d1=0; where_d1<36; where_d1++)
	{
	    ok=1;
	    beta = 1000000;
	    if(IsValidMove(d1_play_b, player, where_d1))
	    {
		pris_d1 = PlayWhite(d1_play_b, d1_calc_b, where_d1);
		Copy(d1_play_b, d2_play_b);
		Copy(d1_calc_b, d2_calc_b);
		for(where_d2=0; ok && where_d2<36; where_d2++)
		{
		    if(IsValidMove(d2_play_b, -1, where_d2))
		    {
			pris_d2 = PlayBlack(d2_play_b, d2_calc_b, where_d2);
			dump = EndScore(d2_play_b, d2_calc_b, player);
			if(dump<beta)
			{
			    beta = dump;
			    if(dump <= alpha)
				ok = 0;
			}
			if(pris_d2)
			{
			    Copy(d1_play_b, d2_play_b);
			    Copy(d1_calc_b, d2_calc_b);
			}
			else
			    UnPlayBlack(d2_play_b, d2_calc_b, where_d2);
		    }
		}
		if (beta>alpha)
		{
		    alpha = beta;
		    where_b = where_d1;
		}
		if(pris_d1)
		{
		    Copy(play_b, d1_play_b);
		    Copy(calc_b, d1_calc_b);
		}
		else
		    UnPlayWhite(d1_play_b, d1_calc_b, where_d1);
		where_result[where_d1] = beta;
	    }
	    else
		where_result[where_d1] = -1000000;
	}
    }
    else		/* Black */
    {
	for(where_d1=0; where_d1<36; where_d1++)
	{
	    ok=1;
	    beta = 1000000;
	    if(IsValidMove(d1_play_b, player, where_d1))
	    {
		pris_d1 = PlayBlack(d1_play_b, d1_calc_b, where_d1);
		Copy(d1_play_b, d2_play_b);
		Copy(d1_calc_b, d2_calc_b);
		for(where_d2=0; ok && where_d2<36; where_d2++)
		{
		    if(IsValidMove(d2_play_b, 1, where_d2))
		    {
			pris_d2 = PlayWhite(d2_play_b, d2_calc_b, where_d2);
			dump = EndScore(d2_play_b, d2_calc_b, player);
			if(dump<beta)
			{
			    beta = dump;
			    if(dump <= alpha)
				ok = 0;
			}
			if(pris_d2)
			{
			    Copy(d1_play_b, d2_play_b);
			    Copy(d1_calc_b, d2_calc_b);
			}
			else
			    UnPlayWhite(d2_play_b, d2_calc_b, where_d2);
		    }
		}
		if (beta>alpha)
		{
		    alpha = beta;
		    where_b = where_d1;
		}
		if(pris_d1)
		{
		    Copy(play_b, d1_play_b);
		    Copy(calc_b, d1_calc_b);
		}
		else
		    UnPlayBlack(d1_play_b, d1_calc_b, where_d1);  
		where_result[where_d1] = beta;
	    }
	    else
		where_result[where_d1] = -10000;
	}
    }
    BubbleSort(where_list, where_result);
    return;
}

static void FastEndShuffler(play_b, calc_b, player, where_list)
    board play_b, calc_b;
    int player, where_list[36];
{
    board d2_play_b, d1_play_b;
    board d2_calc_b, d1_calc_b;
    int alpha, beta;
    int pris_d1, pris_d2, ok, dump;
    int where_d1, where_d2, where_b;
    int where_result[36];
    int st_score;
    param_cell cell;
    alpha = -1000000;
    where_b = -1;

    cell.my_score_rate = 2.0;
    cell.opp_score_rate = 2.0;
    cell.protect_pos_rate = 1.0;
    cell.attack_pos_rate = 1.0;
    cell.board_score_rate = 1.0;
    cell.play_opp_flag = 20.0;
    cell.defensive_rate = 7.0;
    cell.offensive_rate = 0;
    cell.position_rate = 0;

    Copy(play_b, d1_play_b);
    Copy(calc_b, d1_calc_b);

    for(where_d1=0; where_d1<36; where_d1++)
    {
	where_list[where_d1] = where_d1;
	where_result[where_d1] = -1000000;
    }

    if(player>0)	/* White */
    {
	for(where_d1=0; where_d1<36; where_d1++)
	{
	    ok=1;
	    beta = 1000000;
	    if(IsValidMove(d1_play_b, player, where_d1))
	    {
		st_score = (int)StrategicScore(play_b, calc_b, player, where_d1, cell);
		pris_d1 = PlayWhite(d1_play_b, d1_calc_b, where_d1);
		Copy(d1_play_b, d2_play_b);
		Copy(d1_calc_b, d2_calc_b);
		for(where_d2=0; ok && where_d2<36; where_d2++)
		{
		    if(IsValidMove(d2_play_b, -1, where_d2))
		    {
			pris_d2 = PlayBlack(d2_play_b, d2_calc_b, where_d2);
			dump = st_score + EndScore(d2_play_b, d2_calc_b, player);
			if(dump<beta)
			{
			    beta = dump;
			    if(dump <= alpha)
				ok = 0;
			}
			if(pris_d2)
			{
			    Copy(d1_play_b, d2_play_b);
			    Copy(d1_calc_b, d2_calc_b);
			}
			else
			    UnPlayBlack(d2_play_b, d2_calc_b, where_d2);
		    }
		}
		if (beta>alpha)
		{
		    alpha = beta;
		    where_b = where_d1;
		}
		if(pris_d1)
		{
		    Copy(play_b, d1_play_b);
		    Copy(calc_b, d1_calc_b);
		}
		else
		    UnPlayWhite(d1_play_b, d1_calc_b, where_d1);
		where_result[where_d1] = beta;
	    }
	    else
		where_result[where_d1] = -1000000;
	}
    }
    else		/* Black */
    {
	for(where_d1=0; where_d1<36; where_d1++)
	{
	    ok=1;
	    beta = 1000000;
	    if(IsValidMove(d1_play_b, player, where_d1))
	    {
		st_score = (int)StrategicScore(play_b, calc_b, player, where_d1, cell);
		pris_d1 = PlayBlack(d1_play_b, d1_calc_b, where_d1);
		Copy(d1_play_b, d2_play_b);
		Copy(d1_calc_b, d2_calc_b);
		for(where_d2=0; ok && where_d2<36; where_d2++)
		{
		    if(IsValidMove(d2_play_b, 1, where_d2))
		    {
			pris_d2 = PlayWhite(d2_play_b, d2_calc_b, where_d2);
			dump = st_score + EndScore(d2_play_b, d2_calc_b, player);
			if(dump<beta)
			{
			    beta = dump;
			    if(dump <= alpha)
				ok = 0;
			}
			if(pris_d2)
			{
			    Copy(d1_play_b, d2_play_b);
			    Copy(d1_calc_b, d2_calc_b);
			}
			else
			    UnPlayWhite(d2_play_b, d2_calc_b, where_d2);
		    }
		}
		if (beta>alpha)
		{
		    alpha = beta;
		    where_b = where_d1;
		}
		if(pris_d1)
		{
		    Copy(play_b, d1_play_b);
		    Copy(calc_b, d1_calc_b);
		}
		else
		    UnPlayBlack(d1_play_b, d1_calc_b, where_d1);  
		where_result[where_d1] = beta;
	    }
	    else
		where_result[where_d1] = -10000;
	}
    }
    BubbleSort(where_list, where_result);
    return;
}

static PARAM_TYPE PAlphaBetaW(sc_weight, st_weight, play_b, calc_b, depth, num_coup, pwhere, brain, alpha, beta)
    PARAM_TYPE st_weight, sc_weight;
    board play_b, calc_b;
    int depth, num_coup;
    pred pwhere;
    param brain;
    PARAM_TYPE alpha, beta;
{
    board d1_play_b;
    board d1_calc_b;
    PARAM_TYPE dump, score;
    int pris_d1;
    int where_d1, where_b;

    int where_list[36] = {  7, 10, 28, 25, 12, 15, 
			   20, 21,  9,  8, 13, 19,
			   16, 22, 27, 26,  1,  2,
			    3,  4, 11, 17, 23, 29,
			   34, 33, 32, 31, 24, 18,
			   12,  6,  5,  0, 30, 35 };

    where_b = 0;
    score = alpha;

    Copy(play_b, d1_play_b);
    Copy(calc_b, d1_calc_b);
    
    if(depth == 1)
    {
	for(where_d1=0; where_d1<36; where_d1++)
	{
	    if(IsValidMove(d1_play_b, 1, where_list[where_d1]))
	    {
		pris_d1 = PlayWhite(d1_play_b, d1_calc_b, where_list[where_d1]);
		dump = SearchScore(d1_play_b, d1_calc_b, sc_weight, st_weight, 1, brain[num_coup]);
		if(dump>score)
		{
		    score = dump;
		    where_b = where_list[where_d1];
		    if(score >= beta)
			return score;
		}
		if(pris_d1)
		{
		    Copy(play_b, d1_play_b);
		    Copy(calc_b, d1_calc_b);
		}
		else
		    UnPlayWhite(d1_play_b, d1_calc_b, where_list[where_d1]);
	    }
	}
	pwhere[depth] = where_b;
	return score;
    }
    else
    {
	if(depth>2)
	    Shuffler(play_b, calc_b, 1, num_coup, where_list, brain);
	for(where_d1=0; where_d1<36; where_d1++)
	{
	    if(IsValidMove(d1_play_b, 1, where_list[where_d1]))
	    {
		pris_d1 = PlayWhite(d1_play_b, d1_calc_b, where_list[where_d1]);
		dump = -PAlphaBetaB(sc_weight, -st_weight, d1_play_b, d1_calc_b,
				    depth-1, num_coup+1, pwhere, brain, -beta, -score); 
		if(dump>score)
		{
		    score = dump;
		    where_b = where_list[where_d1];
		    if(score >= beta)
			return score;
		}
		if(pris_d1)
		{
		    Copy(play_b, d1_play_b);
		    Copy(calc_b, d1_calc_b);
		}
		else
		    UnPlayWhite(d1_play_b, d1_calc_b, where_list[where_d1]);
	    }
	}
	pwhere[depth] = where_b;
	return score;
    }
}


static PARAM_TYPE PAlphaBetaB(sc_weight, st_weight, play_b, calc_b, depth, num_coup, pwhere, brain, alpha, beta)
    PARAM_TYPE st_weight, sc_weight;
    board play_b, calc_b;
    int depth, num_coup;
    pred pwhere;
    param brain;
    PARAM_TYPE alpha, beta;
{
    board d1_play_b;
    board d1_calc_b;
    PARAM_TYPE dump, score;
    int pris_d1;
    int where_d1, where_b;

    int where_list[36] = {  7, 10, 28, 25, 12, 15, 
			   20, 21,  9,  8, 13, 19,
			   16, 22, 27, 26,  1,  2,
			    3,  4, 11, 17, 23, 29,
			   34, 33, 32, 31, 24, 18,
			   12,  6,  5,  0, 30, 35 };

    where_b = 0;
    score = alpha;

    Copy(play_b, d1_play_b);
    Copy(calc_b, d1_calc_b);
    
    if(depth == 1)
    {
	for(where_d1=0; where_d1<36; where_d1++)
	{
	    if(IsValidMove(d1_play_b, -1, where_list[where_d1]))
	    {
		pris_d1 = PlayBlack(d1_play_b, d1_calc_b, where_list[where_d1]);
		dump = SearchScore(d1_play_b, d1_calc_b, sc_weight, st_weight, -1, brain[num_coup]);
		if(dump>score)
		{
		    score = dump;
		    where_b = where_list[where_d1];
		    if(score >= beta)
			return score;
		}
		if(pris_d1)
		{
		    Copy(play_b, d1_play_b);
		    Copy(calc_b, d1_calc_b);
		}
		else
		    UnPlayBlack(d1_play_b, d1_calc_b, where_list[where_d1]);
	    }
	}
	pwhere[depth] = where_b;
	return score;
    }
    else
    {
	if(depth>2)
	    Shuffler(play_b, calc_b, -1, num_coup, where_list, brain);
	for(where_d1=0; where_d1<36; where_d1++)
	{
	    if(IsValidMove(d1_play_b, -1, where_list[where_d1]))
	    {
		pris_d1 = PlayBlack(d1_play_b, d1_calc_b, where_list[where_d1]);
		dump = -PAlphaBetaW(sc_weight, -st_weight, d1_play_b, d1_calc_b,
				    depth-1, num_coup+1, pwhere, brain, -beta, -score); 
		if(dump>score)
		{
		    score = dump;
		    where_b = where_list[where_d1];
		    if(score >= beta)
			return score;
		}
		if(pris_d1)
		{
		    Copy(play_b, d1_play_b);
		    Copy(calc_b, d1_calc_b);
		}
		else
		    UnPlayBlack(d1_play_b, d1_calc_b, where_list[where_d1]);
	    }
	}
	pwhere[depth] = where_b; /* not a cut-off update list */
	return score;   
    }
}



int AlphaBeta(play_b, calc_b, player, depth, num_coup, pwhere, brain)
    board play_b, calc_b;
    int player, depth, num_coup;
    pred pwhere;
    param brain;
{
    board d1_play_b;
    board d1_calc_b;
    PARAM_TYPE dump, sc_weight, st_weight, alpha, beta;
    int pris_d1;
    int where_d1, where_b;

    int where_list[36] = {  7, 10, 28, 25, 12, 15, 
			   20, 21,  9,  8, 13, 19,
			   16, 22, 27, 26,  1,  2,
			    3,  4, 11, 17, 23, 29,
			   34, 33, 32, 31, 24, 18,
			   12,  6,  5,  0, 30, 35 };

    where_b = 0;
    alpha = -1000000;
    beta = 1000000;

    sc_weight = brain[num_coup].board_score_rate;
    
    Copy(play_b, d1_play_b);
    Copy(calc_b, d1_calc_b);
    
    if(depth == 1)
    {
	if(player>0)	/* White */
	{
	    for(where_d1=0; where_d1<36; where_d1++)
	    {
		if(IsValidMove(d1_play_b, player, where_list[where_d1]))
		{
		    st_weight = StrategicScore(play_b, calc_b, player, where_list[where_d1], brain[num_coup]);
		    pris_d1 = PlayWhite(d1_play_b, d1_calc_b, where_list[where_d1]);
		    dump = SearchScore(d1_play_b, d1_calc_b, sc_weight, st_weight, player, brain[num_coup]);
		    if(dump>alpha)
		    {
			alpha = dump;
			where_b = where_list[where_d1];
		    }
		    if(pris_d1)
		    {
			Copy(play_b, d1_play_b);
			Copy(calc_b, d1_calc_b);
		    }
		    else
			UnPlayWhite(d1_play_b, d1_calc_b, where_list[where_d1]);
		}
	    }
	    pwhere[depth] = where_b;
	    return alpha;
	}
	else
	{			/* Black */
	    for(where_d1=0; where_d1<36; where_d1++)
	    {
		if(IsValidMove(d1_play_b, player, where_list[where_d1]))
		{
		    st_weight = StrategicScore(play_b, calc_b, player, where_list[where_d1], brain[num_coup]);
		    pris_d1 = PlayBlack(d1_play_b, d1_calc_b, where_list[where_d1]);
		    dump = SearchScore(d1_play_b, d1_calc_b, sc_weight, st_weight, player, brain[num_coup]);
		    if(dump>alpha)
		    {
			alpha = dump;
			where_b = where_list[where_d1];
		    }
		    if(pris_d1)
		    {
			Copy(play_b, d1_play_b);
			Copy(calc_b, d1_calc_b);
		    }
		    else
			UnPlayBlack(d1_play_b, d1_calc_b, where_list[where_d1]);
		}
	    }
	    pwhere[depth] = where_b;
	    return alpha;
	}
    }
    else /* propagation */
    {
#ifdef ITERATIVE_WINDOW
	if(depth>4)
	{
	    dump = AlphaBeta(play_b, calc_b, player, depth-2, num_coup, pwhere, brain);
	    alpha = dump-4;
	    beta = dump+4;
	}
#endif /* ITERATIVE_WINDOW */
	Shuffler(play_b, calc_b, player, num_coup, where_list, brain);
	if(player>0) /* White */
	{
	    for(where_d1=0; where_d1<36; where_d1++)
	    {
		if(IsValidMove(d1_play_b, 1, where_list[where_d1]))
		{
		    st_weight = StrategicScore(play_b, calc_b, player, where_list[where_d1], brain[num_coup]);
		    pris_d1 = PlayWhite(d1_play_b, d1_calc_b, where_list[where_d1]);
		    dump = -PAlphaBetaB(sc_weight, -st_weight, d1_play_b, d1_calc_b,
					depth-1, num_coup+1, pwhere, brain, -beta, -alpha); 
		    if(dump>alpha)
		    {
			alpha = dump;
			where_b = where_list[where_d1];
		    }
		    if(pris_d1)
		    {
			Copy(play_b, d1_play_b);
			Copy(calc_b, d1_calc_b);
		    }
		    else
			UnPlayWhite(d1_play_b, d1_calc_b, where_list[where_d1]);
		}
	    }
	    pwhere[depth] = where_b;
	    return alpha;
	}
  	else /* Black */
	{
	    for(where_d1=0; where_d1<36; where_d1++)
	    {
		if(IsValidMove(d1_play_b, player, where_list[where_d1]))
		{
		    st_weight = StrategicScore(play_b, calc_b, player, where_list[where_d1], brain[num_coup]);
		    pris_d1 = PlayBlack(d1_play_b, d1_calc_b, where_list[where_d1]);
		    dump = -PAlphaBetaW(sc_weight, -st_weight, d1_play_b, d1_calc_b,
					depth-1, num_coup+1, pwhere, brain, -beta, -alpha); 
		    if(dump>alpha)
		    {
			alpha = dump;
			where_b = where_list[where_d1];
		    }
		    if(pris_d1)
		    {
			Copy(play_b, d1_play_b);
			Copy(calc_b, d1_calc_b);
		    }
		    else
			UnPlayBlack(d1_play_b, d1_calc_b, where_list[where_d1]);
		}
	    }
	    pwhere[depth] = where_b;
	    return alpha;
	}
    }
}
	

static int PEndAlphaBetaW(play_b, calc_b, depth, num_coup, pwhere, rwhere, alpha, beta)
    board play_b, calc_b;
    int depth, num_coup;
    pred pwhere;
    pred rwhere;
    int alpha, beta;
{
    board d1_play_b;
    board d1_calc_b;
    int dump, score, exc;
    int pris_d1, i;
    int where_d1, where_b;
    pred best;
    int where_list[36] = {  0,  1,  2,  3,  4,  5,
			   11, 17, 23, 29, 35, 34,
			   33, 32, 31, 30, 24, 18,
			   12,  6, 14, 15, 20, 21,
			   28, 27, 26, 25, 19, 13,
			    7,  8,  9, 10, 16, 22 };

    where_b = 0;
    score = alpha;

    Copy(play_b, d1_play_b);
    Copy(calc_b, d1_calc_b);
    
    if(depth == 1)
    {
	for(where_d1=0; where_d1<36; where_d1++)
	{
	    if(IsValidMove(d1_play_b, 1, where_list[where_d1]))
	    {
		pris_d1 = PlayWhite(d1_play_b, d1_calc_b, where_list[where_d1]);
		dump = EndScore(d1_play_b, d1_calc_b, 1);
		if(dump>score)
		{
		    score = dump;
		    where_b = where_list[where_d1];
		    if(score >= beta)
			return score;
		}
		if(pris_d1)
		{
		    Copy(play_b, d1_play_b);
		    Copy(calc_b, d1_calc_b);
		}
		else
		    UnPlayWhite(d1_play_b, d1_calc_b, where_list[where_d1]);
	    }
	}
	pwhere[depth] = where_b;
	return score;
    }
    else
    {
	if(depth>2)
	    EndShuffler(play_b, calc_b, 1, where_list);
	
	if(rwhere[depth] != -1) /* if recorded route , reorganisation */
	{
	    for(dump = rwhere[depth], i=0; where_list[i] != rwhere[depth]; i++)
	    {
		exc = where_list[i];
		where_list[i] = dump;
		dump = exc;
	    }
	    where_list[i] = dump;
	}

	for(where_d1=0; where_d1<36; where_d1++)
	{
	    if(IsValidMove(d1_play_b, 1, where_list[where_d1]))
	    {
		pris_d1 = PlayWhite(d1_play_b, d1_calc_b, where_list[where_d1]);
		dump = -PEndAlphaBetaB(d1_play_b, d1_calc_b, depth-1, num_coup+1,
				       pwhere, rwhere, -beta, -score); 
		if(dump>score)
		{
		    score = dump;
		    where_b = where_list[where_d1];
		    if(score >= beta)
			return score;
		    for(i=1; i<depth; i++)
			best[i] = pwhere[i];
		}
		if(pris_d1)
		{
		    Copy(play_b, d1_play_b);
		    Copy(calc_b, d1_calc_b);
		}
		else
		    UnPlayWhite(d1_play_b, d1_calc_b, where_list[where_d1]);
	    }
	}
	pwhere[depth] = where_b;
	for(i=1; i<depth; i++)
	    pwhere[i] = best[i];
	return score;
    }
}


static int PEndAlphaBetaB(play_b, calc_b, depth, num_coup, pwhere, rwhere, alpha, beta)
    board play_b, calc_b;
    int depth, num_coup;
    pred pwhere;
    pred rwhere;
    int alpha, beta;
{
    board d1_play_b;
    board d1_calc_b;
    int dump, score, exc;
    int pris_d1, i;
    int where_d1, where_b;
    pred best;
    int where_list[36] = {  0,  1,  2,  3,  4,  5,
			   11, 17, 23, 29, 35, 34,
			   33, 32, 31, 30, 24, 18,
			   12,  6, 14, 15, 20, 21,
			   28, 27, 26, 25, 19, 13,
			    7,  8,  9, 10, 16, 22 };

  
    where_b = 0;
    score = alpha;

    Copy(play_b, d1_play_b);
    Copy(calc_b, d1_calc_b);
    
    if(depth == 1)
    {
	for(where_d1=0; where_d1<36; where_d1++)
	{
	    if(IsValidMove(d1_play_b, -1, where_list[where_d1]))
	    {
		pris_d1 = PlayBlack(d1_play_b, d1_calc_b, where_list[where_d1]);
		dump = EndScore(d1_play_b, d1_calc_b, -1);
		if(dump>score)
		{
		    score = dump;
		    where_b = where_list[where_d1];
		    if(score >= beta)
			return score;
		}
		if(pris_d1)
		{
		    Copy(play_b, d1_play_b);
		    Copy(calc_b, d1_calc_b);
		}
		else
		    UnPlayBlack(d1_play_b, d1_calc_b, where_list[where_d1]);
	    }
	}
	pwhere[depth] = where_b;
	return score;
    }
    else
    {
	if(depth>2)
	    EndShuffler(play_b, calc_b, -1, where_list);

	if(rwhere[depth] != -1) /* if recorded route , reorganisation */
	{
	    for(dump = rwhere[depth], i=0; where_list[i] != rwhere[depth]; i++)
	    {
		exc = where_list[i];
		where_list[i] = dump;
		dump = exc;
	    }
	    where_list[i] = dump;
	}
    
	for(where_d1=0; where_d1<36; where_d1++)
	{
	    if(IsValidMove(d1_play_b, -1, where_list[where_d1]))
	    {
		pris_d1 = PlayBlack(d1_play_b, d1_calc_b, where_list[where_d1]);
		dump = -PEndAlphaBetaW(d1_play_b, d1_calc_b, depth-1, num_coup+1,
				       pwhere, rwhere, -beta, -score); 
		if(dump>score)
		{
		    score = dump;
		    where_b = where_list[where_d1];
		    if(score >= beta)
			return score;
		    for(i=1; i<depth; i++)
			best[i] = pwhere[i];
		}
		if(pris_d1)
		{
		    Copy(play_b, d1_play_b);
		    Copy(calc_b, d1_calc_b);
		}
		else
		    UnPlayBlack(d1_play_b, d1_calc_b, where_list[where_d1]);
	    }
	}
	pwhere[depth] = where_b; /* not a cut-off update list */
	for(i=1; i<depth; i++)
	    pwhere[i] = best[i];
	return score;   
    }
}


int EndAlphaBeta(play_b, calc_b, player, depth, num_coup, pwhere)
    board play_b, calc_b;
    int player, depth, num_coup;
    pred pwhere;
{
    board d1_play_b;
    board d1_calc_b;
    int dump, alpha, beta;
    int pris_d1, i, exc;
    int where_d1, where_b;
    pred recorded = { -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
    pred best;

    int where_list[36] = {  0,  1,  2,  3,  4,  5,
			     11, 17, 23, 29, 35, 34,
			     33, 32, 31, 30, 24, 18,
			     12,  6, 14, 15, 20, 21,
			     28, 27, 26, 25, 19, 13,
			      7,  8,  9, 10, 16, 22 };

    where_b = 0;
    alpha = -1000000;
    beta = 1000000;

    Copy(play_b, d1_play_b);
    Copy(calc_b, d1_calc_b);
    
    if(depth == 1)
    {
	if(player>0)	/* White */
	{
	    for(where_d1=0; where_d1<36; where_d1++)
	    {
		if(IsValidMove(d1_play_b, player, where_list[where_d1]))
		{
		    pris_d1 = PlayWhite(d1_play_b, d1_calc_b, where_list[where_d1]);
		    dump = EndScore(d1_play_b, d1_calc_b, player);
		    if(dump>alpha)
		    {
			alpha = dump;
			where_b = where_list[where_d1];
		    }
		    if(pris_d1)
		    {
			Copy(play_b, d1_play_b);
			Copy(calc_b, d1_calc_b);
		    }
		    else
			UnPlayWhite(d1_play_b, d1_calc_b, where_list[where_d1]);
		}
	    }
	    pwhere[depth] = where_b;
	    return alpha;
	}
	else
	{			/* Black */
	    for(where_d1=0; where_d1<36; where_d1++)
	    {
		if(IsValidMove(d1_play_b, player, where_list[where_d1]))
		{
		    pris_d1 = PlayBlack(d1_play_b, d1_calc_b, where_list[where_d1]);
		    dump = EndScore(d1_play_b, d1_calc_b, player);
		    if(dump>alpha)
		    {
			alpha = dump;
			where_b = where_list[where_d1];
		    }
		    if(pris_d1)
		    {
			Copy(play_b, d1_play_b);
			Copy(calc_b, d1_calc_b);
		    }
		    else
			UnPlayBlack(d1_play_b, d1_calc_b, where_list[where_d1]);
		}
	    }
	    pwhere[depth] = where_b;
	    return alpha;
	}
    }
    else /* propagation */
    {
	if(depth>3) /* iterative deepening */
	{
	    dump = EndAlphaBeta(play_b, calc_b, player, depth-STEP, num_coup, pwhere);
	    for(i=1; i<depth-(STEP-1); i++)
		recorded[i+STEP] = pwhere[i];
#ifdef USE_WINDOW
	    alpha = dump-1;
	    beta = dump+1;
#endif /* USE_WINDOW */
	}
	EndShuffler(play_b, calc_b, player, where_list);
	if(recorded[depth] != -1)
	{
	    for(dump = recorded[depth], i=0; where_list[i] != recorded[depth]; i++)
	    {
		exc = where_list[i];
		where_list[i] = dump;
		dump = exc;
	    }
	    where_list[i] = dump;
	}
	if(player>0) /* White */
	{
	    for(where_d1=0; where_d1<36; where_d1++)
	    {
		if(IsValidMove(d1_play_b, 1, where_list[where_d1]))
		{
		    pris_d1 = PlayWhite(d1_play_b, d1_calc_b, where_list[where_d1]);
		    dump = -PEndAlphaBetaB(d1_play_b, d1_calc_b, depth-1, num_coup+1,
					   pwhere, recorded, -beta, -alpha); 
		    if(dump>alpha)
		    {
			alpha = dump;
			where_b = where_list[where_d1];
			for(i=1; i<depth; i++)
			    best[i] = pwhere[i];
		    }
		    if(pris_d1)
		    {
			Copy(play_b, d1_play_b);
			Copy(calc_b, d1_calc_b);
		    }
		    else
			UnPlayWhite(d1_play_b, d1_calc_b, where_list[where_d1]);
		}
	    }
	    pwhere[depth] = where_b;
	    for(i=1; i<depth; i++)
		pwhere[i] = best[i];
#ifdef VERBOSE_ROUTE
	    for(i=0;i<depth;i++)
		printf("[%d]", pwhere[i+1]);
	    printf("\n");
#endif /* VERBOSE_ROUTE */
	    return alpha;
	}
  	else /* Black */
	{
	    for(where_d1=0; where_d1<36; where_d1++)
	    {
		if(IsValidMove(d1_play_b, player, where_list[where_d1]))
		{
		    pris_d1 = PlayBlack(d1_play_b, d1_calc_b, where_list[where_d1]);
		    dump = -PEndAlphaBetaW(d1_play_b, d1_calc_b, depth-1, num_coup+1,
					   pwhere, recorded, -beta, -alpha); 
		    if(dump>alpha)
		    {
			alpha = dump;
			where_b = where_list[where_d1];
		   	for(i=1; i<depth; i++)
			    best[i] = pwhere[i];
		    }
		    if(pris_d1)
		    {
			Copy(play_b, d1_play_b);
			Copy(calc_b, d1_calc_b);
		    }
		    else
			UnPlayBlack(d1_play_b, d1_calc_b, where_list[where_d1]);
		}
	    }
	    pwhere[depth] = where_b;
	    for(i=1; i<depth; i++)
		pwhere[i] = best[i];
#ifdef VERBOSE_ROUTE
	    for(i=0;i<depth;i++)
		printf("[%d]", pwhere[i+1]);
	    printf("\n");
#endif /* VERBOSE_ROUTE */
	    return alpha;
	}
    }
}



static int PFastEndAlphaBetaW(play_b, calc_b, depth, num_coup, pwhere, rwhere, alpha, beta)
    board play_b, calc_b;
    int depth, num_coup;
    pred pwhere, rwhere;
    int alpha, beta;
{
    board d1_play_b;
    board d1_calc_b;
    int dump, score, exc;
    int pris_d1, i;
    int where_d1, where_b;
    pred best;
    int where_list[36] = {  0,  1,  2,  3,  4,  5,
			   11, 17, 23, 29, 35, 34,
			   33, 32, 31, 30, 24, 18,
			   12,  6, 14, 15, 20, 21,
			   28, 27, 26, 25, 19, 13,
			    7,  8,  9, 10, 16, 22 };

    where_b = 0;
    score = alpha;

    Copy(play_b, d1_play_b);
    Copy(calc_b, d1_calc_b);
    
    if(depth == 1)
    {
	for(where_d1=0; where_d1<36; where_d1++)
	{
	    if(IsValidMove(d1_play_b, 1, where_list[where_d1]))
	    {
		pris_d1 = PlayWhite(d1_play_b, d1_calc_b, where_list[where_d1]);
		dump = EndScore(d1_play_b, d1_calc_b, 1);
		if(dump>score)
		{
		    score = dump;
		    where_b = where_list[where_d1];
		    if(score >= beta)
			return score;
		}
		if(pris_d1)
		{
		    Copy(play_b, d1_play_b);
		    Copy(calc_b, d1_calc_b);
		}
		else
		    UnPlayWhite(d1_play_b, d1_calc_b, where_list[where_d1]);
	    }
	}
	pwhere[depth] = where_b;
	return score;
    }
    else
    {
	if(depth>2)
	    FastEndShuffler(play_b, calc_b, 1, where_list);
	if(rwhere[depth] != -1) /* if recorded route , reorganisation */
	{
	    for(dump = rwhere[depth], i=0; where_list[i] != rwhere[depth]; i++)
	    {
		exc = where_list[i];
		where_list[i] = dump;
		dump = exc;
	    }
	    where_list[i] = dump;
	}
	for(where_d1=0; where_d1<IHARDCUT; where_d1++)
	{
	    if(IsValidMove(d1_play_b, 1, where_list[where_d1]))
	    {
		pris_d1 = PlayWhite(d1_play_b, d1_calc_b, where_list[where_d1]);
		dump = -PFastEndAlphaBetaB(d1_play_b, d1_calc_b, depth-1, num_coup+1,
					   pwhere, rwhere, -beta, -score); 
		if(dump>score)
		{
		    score = dump;
		    where_b = where_list[where_d1];
		    if(score >= beta)
			return score;
		    for(i=1; i<depth; i++)
			best[i] = pwhere[i];
		}
		if(pris_d1)
		{
		    Copy(play_b, d1_play_b);
		    Copy(calc_b, d1_calc_b);
		}
		else
		    UnPlayWhite(d1_play_b, d1_calc_b, where_list[where_d1]);
	    }
	}
	pwhere[depth] = where_b;
	for(i=1; i<depth; i++)
	    pwhere[i] = best[i];
	return score;
    }
}


static int PFastEndAlphaBetaB(play_b, calc_b, depth, num_coup, pwhere, rwhere, alpha, beta)
    board play_b, calc_b;
    int depth, num_coup;
    pred pwhere, rwhere;
    int alpha, beta;
{
    board d1_play_b;
    board d1_calc_b;
    int dump, score, exc;
    int pris_d1, i;
    int where_d1, where_b;
    pred best;
    int where_list[36] = {  0,  1,  2,  3,  4,  5,
			   11, 17, 23, 29, 35, 34,
			   33, 32, 31, 30, 24, 18,
			   12,  6, 14, 15, 20, 21,
			   28, 27, 26, 25, 19, 13,
			    7,  8,  9, 10, 16, 22 };

  
    where_b = 0;
    score = alpha;

    Copy(play_b, d1_play_b);
    Copy(calc_b, d1_calc_b);
    
    if(depth == 1)
    {
	for(where_d1=0; where_d1<36; where_d1++)
	{
	    if(IsValidMove(d1_play_b, -1, where_list[where_d1]))
	    {
		pris_d1 = PlayBlack(d1_play_b, d1_calc_b, where_list[where_d1]);
		dump = EndScore(d1_play_b, d1_calc_b, -1);
		if(dump>score)
		{
		    score = dump;
		    where_b = where_list[where_d1];
		    if(score >= beta)
			return score;
		}
		if(pris_d1)
		{
		    Copy(play_b, d1_play_b);
		    Copy(calc_b, d1_calc_b);
		}
		else
		    UnPlayBlack(d1_play_b, d1_calc_b, where_list[where_d1]);
	    }
	}
	pwhere[depth] = where_b;
	return score;
    }
    else
    {
	if(depth>2)
	    FastEndShuffler(play_b, calc_b, -1, where_list);
	if(rwhere[depth] != -1) /* if recorded route , reorganisation */
	{
	    for(dump = rwhere[depth], i=0; where_list[i] != rwhere[depth]; i++)
	    {
		exc = where_list[i];
		where_list[i] = dump;
		dump = exc;
	    }
	    where_list[i] = dump;
	}
	for(where_d1=0; where_d1<IHARDCUT; where_d1++)
	{
	    if(IsValidMove(d1_play_b, -1, where_list[where_d1]))
	    {
		pris_d1 = PlayBlack(d1_play_b, d1_calc_b, where_list[where_d1]);
		dump = -PFastEndAlphaBetaW(d1_play_b, d1_calc_b, depth-1, num_coup+1,
					   pwhere, rwhere, -beta, -score); 
		if(dump>score)
		{
		    score = dump;
		    where_b = where_list[where_d1];
		    if(score >= beta)
			return score;
		    for(i=1; i<depth; i++)
			best[i] = pwhere[i];
		}
		if(pris_d1)
		{
		    Copy(play_b, d1_play_b);
		    Copy(calc_b, d1_calc_b);
		}
		else
		    UnPlayBlack(d1_play_b, d1_calc_b, where_list[where_d1]);
	    }
	}
	pwhere[depth] = where_b; /* not a cut-off update list */
	for(i=1; i<depth; i++)
	    pwhere[i] = best[i];
	return score;   
    }
}



int FastEndAlphaBeta(play_b, calc_b, player, depth, num_coup, pwhere)
    board play_b, calc_b;
    int player, depth, num_coup;
    pred pwhere;
{
    board d1_play_b;
    board d1_calc_b;
    int dump, alpha, beta, exc;
    int pris_d1, i;
    int where_d1, where_b;
    pred recorded = {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
    pred best;
    int where_list[36] = {  0,  1,  2,  3,  4,  5,
			     11, 17, 23, 29, 35, 34,
			     33, 32, 31, 30, 24, 18,
			     12,  6, 14, 15, 20, 21,
			     28, 27, 26, 25, 19, 13,
			      7,  8,  9, 10, 16, 22 };

    where_b = 0;
    alpha = -1000000;
    beta = 1000000;

    Copy(play_b, d1_play_b);
    Copy(calc_b, d1_calc_b);
    
    if(depth == 1)
    {
	if(player>0)	/* White */
	{
	    for(where_d1=0; where_d1<36; where_d1++)
	    {
		if(IsValidMove(d1_play_b, player, where_list[where_d1]))
		{
		    pris_d1 = PlayWhite(d1_play_b, d1_calc_b, where_list[where_d1]);
		    dump = EndScore(d1_play_b, d1_calc_b, player);
		    if(dump>alpha)
		    {
			alpha = dump;
			where_b = where_list[where_d1];
		    }
		    if(pris_d1)
		    {
			Copy(play_b, d1_play_b);
			Copy(calc_b, d1_calc_b);
		    }
		    else
			UnPlayWhite(d1_play_b, d1_calc_b, where_list[where_d1]);
		}
	    }
	    pwhere[depth] = where_b;
	    return alpha;
	}
	else
	{			/* Black */
	    for(where_d1=0; where_d1<36; where_d1++)
	    {
		if(IsValidMove(d1_play_b, player, where_list[where_d1]))
		{
		    pris_d1 = PlayBlack(d1_play_b, d1_calc_b, where_list[where_d1]);
		    dump = EndScore(d1_play_b, d1_calc_b, player);
		    if(dump>alpha)
		    {
			alpha = dump;
			where_b = where_list[where_d1];
		    }
		    if(pris_d1)
		    {
			Copy(play_b, d1_play_b);
			Copy(calc_b, d1_calc_b);
		    }
		    else
			UnPlayBlack(d1_play_b, d1_calc_b, where_list[where_d1]);
		}
	    }
	    pwhere[depth] = where_b;
	    return alpha;
	}
    }
    else /* propagation */
    {
	if(depth>3) /* at an inferior depth, EndAlphaBeta is ok to use */
	{
	    dump = FastEndAlphaBeta(play_b, calc_b, player, depth-STEP, num_coup, pwhere);
	    for(i=1; i<depth-(STEP-1); i++)
		recorded[i+STEP] = pwhere[i];
#ifdef USE_WINDOW
	    alpha = dump-1;
	    beta = dump+1;
#endif /* USE_WINDOW */
	}

	FastEndShuffler(play_b, calc_b, player, where_list);
	if(recorded[depth] != -1)
	{
	    for(dump = recorded[depth], i=0; where_list[i] != recorded[depth]; i++)
	    {
		exc = where_list[i];
		where_list[i] = dump;
		dump = exc;
	    }
	    where_list[i] = dump;
	}
	if(player>0) /* White */
	{
	    for(where_d1=0; where_d1<HARDCUT; where_d1++)
	    {
		if(IsValidMove(d1_play_b, 1, where_list[where_d1]))
		{
		    pris_d1 = PlayWhite(d1_play_b, d1_calc_b, where_list[where_d1]);
		    dump = -PFastEndAlphaBetaB(d1_play_b, d1_calc_b, depth-1, num_coup+1,
					       pwhere, recorded, -beta, -alpha); 
		    if(dump>alpha)
		    {
			alpha = dump;
			where_b = where_list[where_d1];
			for(i=1; i<depth; i++)
			    best[i] = pwhere[i];
		    }
		    if(pris_d1)
		    {
			Copy(play_b, d1_play_b);
			Copy(calc_b, d1_calc_b);
		    }
		    else
			UnPlayWhite(d1_play_b, d1_calc_b, where_list[where_d1]);
		}
	    }
	    pwhere[depth] = where_b;
	    for(i=1; i<depth; i++)
		pwhere[i] = best[i];
#ifdef VERBOSE_ROUTE
	    for(i=0;i<depth;i++)
		printf("[%d]", pwhere[i+1]);
	    printf("\n");
#endif /* VERBOSE_ROUTE */
	    return alpha;
	}
  	else /* Black */
	{
	    for(where_d1=0; where_d1<HARDCUT; where_d1++)
	    {
		if(IsValidMove(d1_play_b, player, where_list[where_d1]))
		{
		    pris_d1 = PlayBlack(d1_play_b, d1_calc_b, where_list[where_d1]);
		    dump = -PFastEndAlphaBetaW(d1_play_b, d1_calc_b, depth-1, num_coup+1,
					       pwhere, recorded, -beta, -alpha); 
		    if(dump>alpha)
		    {
			alpha = dump;
			where_b = where_list[where_d1];
			for(i=1; i<depth; i++)
			    best[i] = pwhere[i];
		    }
		    if(pris_d1)
		    {
			Copy(play_b, d1_play_b);
			Copy(calc_b, d1_calc_b);
		    }
		    else
			UnPlayBlack(d1_play_b, d1_calc_b, where_list[where_d1]);
		}
	    }
	    pwhere[depth] = where_b;
	    for(i=1; i<depth; i++)
		pwhere[i] = best[i];
#ifdef VERBOSE_ROUTE
	    for(i=0;i<depth;i++)
		printf("[%d]", pwhere[i+1]);
	    printf("\n");
#endif /* VERBOSE_ROUTE */
	    return alpha;
	}
    }
}

